/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file include/applications/afs/encoding/block.h
 * @author Christian Grothoff
 **/

#ifndef BLOCK_H
#define BLOCK_H

#include "util/semaphore.h"
#include "util/tcpio.h"
#include "util/cron.h"
#include "util/hashing.h"

/* size of the Blocks we slice file data into 
   (DBlocks and IBlocks). Never change this! */
#define CONTENT_SIZE 1024

/**
 * A CONTENT_Block, representative of the structure
 * of the leaf nodes (a simple chunk of 1 kb of data)
 **/
typedef struct {
  unsigned char content[CONTENT_SIZE]; 
} CONTENT_Block;

/**
 * Pair of Hashcodes for CHK encoded blocks.
 * Every DBlock and IBlock is represented by two
 * hashcodes, one is the key used to encrypt or
 * decrypt the block; the other one is used to
 * search for the block without reveiling the key.
 * See also Freenet's CHK keys.<p>
 * 
 * Note that GNUnet uses a different encoding for
 * the RBlocks (root-nodes) in order to make searches
 * possible.
 **/
typedef struct {

  /**
   * The hash of the plaintext is the key to decrypt.
   **/
  HashCode160 key;

  /**
   * The hash of the encrypted block is the query.
   **/
  HashCode160 query;

} CHK_Hashes;

/* ********* IOContext for encapsulation of IO ********** */

/**
 * IO context for reading-writing file blocks.<p>
 *
 * In GNUnet, files are stored in the form of a balanced tree, not
 * unlike INodes in unix filesystems. When we download files, the
 * inner nodes of the tree are stored under FILENAME.X (where X
 * characterizes the level of the node in the tree). If the download
 * is aborted and resumed later, these .X files can be used to avoid
 * downloading the inner blocks again.  The successfully received leaf
 * nodes in FILENAME (the target file) are of course also not
 * downloaded again.<p>
 *
 * The IOContext struct presents an easy api to access the various
 * dot-files. It uses function pointers to allow implementors to
 * provide a different mechanism (other than files on the drive) to
 * cache the IBlocks.
 **/
typedef struct IOContext {

  /**
   * The depth of the file-tree.
   **/
  int treedepth;
  
  /**
   * A lock for each file-handle for synchronizing access.
   **/
  Mutex * locks;

  /**
   * The file handles for each level in the tree.
   **/
  int * handles;

  /**
   * The base-filename
   **/
  char * filename;

} IOContext;

/**
 * Initialize an IOContext.
 *
 * @param this the context to initialize
 * @param filesize the size of the file
 * @param filename the name of the level-0 file
 * @param rdOnly use YES for read-only IOC
 * @return OK on success, SYSERR on failure
 **/
int createIOContext(IOContext * this,
		    size_t filesize,
		    char * filename,
		    int rdOnly);

/**
 * Read method.
 *
 * @param this reference to the IOContext
 * @param level level in the tree to read/write at
 * @param size_t position where to read or write
 * @param buf where to read from or write to
 * @param len how many bytes to read or write
 * @return number of bytes read or written, SYSERR on error  
 **/
int readFromIOC(struct IOContext * this,
		int level,
		size_t pos,
		void * buf,
		int len);

/**
 * Write method.
 *
 * @param this reference to the IOContext
 * @param level level in the tree to read/write at
 * @param size_t position where to read or write
 * @param buf where to read from or write to
 * @param len how many bytes to read or write
 * @return number of bytes read or written, SYSERR on error  
 **/
int writeToIOC(struct IOContext * this,
	       int level,
	       size_t pos,
	       void * buf,
	       int len);

/**
 * Close the files in the IOContext and free
 * the associated resources. Does NOT free
 * the memory occupied by the IOContext struct
 * itself.
 *
 * @param this reference to the IOContext
 * @param unlinkTreeFiles if YES, the non-level 0 files
 *     are unlinked (removed), set to NO if the download
 *     is not complete and may be resumed later.
 **/
void freeIOC(struct IOContext * this,
	     int unlinkTreeFiles);

/* ******************* the request manager ************* */

/**
 * How many requests do we have pending (max)
 * Must be > 50 to make sure we trigger flushing buffers
 * at the node. Lower values should increase anonymity
 * and decrease performance (for the individual node
 * that makes this change).
 **/
#define MAX_REQUESTS 256

/**
 * Handle to the state of a request manager.
 **/
typedef struct RequestManager {

  /**
   * Mutex for synchronizing access to this struct
   **/
  Mutex lock;

  /**
   * Current list of all pending requests
   **/ 
  struct RequestEntry ** requestList;

  /**
   * Number of pending requests (highest used index)
   **/
  int requestListIndex;

  /**
   * Number of entries allocated for requestList
   **/
  int requestListSize;

  /**
   * Current list of all requests that have been send out
   **/ 
  struct RequestEntry * requestedList[MAX_REQUESTS];

  /**
   * Number of pending requests (highest index)
   **/
  int requestedListIndex;

  /**
   * Current "good" TTL (initial) [64s]
   **/
  unsigned int initialTTL;

  /**
   * Current estimate of "duplication" rate (amount of
   * duplicate replies we get).
   **/
  int duplicationEstimate;

  /**
   * Socket used to talk to gnunetd.
   **/
  GNUNET_TCP_SOCKET * sock;

  /**
   * The thread that receives results from gnunetd.
   **/
  Semaphore * destroySignal;

} RequestManager;


/**
 * TCP communication: search result content send back by gnunetd
 **/
typedef struct {

  /**
   * The TCP header (values: sizeof(AFS_CS_RESULT_CHK), AFS_CS_PROTO_RESULT_CHK) 
   **/ 
  CS_HEADER header;

  /**
   * The search result.
   **/
  CONTENT_Block result;

} AFS_CS_RESULT_CHK;

/**
 * Type of a method that is called by the RequestManager
 * whenever a reply to a query has been received.
 *
 * @param this the node in the tree for which the request
 *       was issued
 * @param query the query that was sent out
 * @param reply the reply that was received
 * @param rm the handle for the request manager
 * @param data an opaque handle that is passed along,
 *        typically used to pass the NodeContext
 * @return SYSERR the request manager should abort the download
 **/
typedef int (*Listener)(void * this,
			HashCode160 * query,
			AFS_CS_RESULT_CHK * reply,
			struct RequestManager * rm,
			void * data); 


/**
 * Message for a set of queries.
 **/
typedef struct {

  /** 
   * The TCP header (values: sizeof(AFS_CS_QUERY), AFS_CS_PROTO_QUERY) 
   **/ 
  CS_HEADER header; 

  /**
   * how important is this request (network byte order) 
   **/  
  unsigned int priority; 

  /**
   * time to live in cronMILLIS (network byte order) 
   **/
  int ttl;      

  /**
   * Hashcodes of the file(s) we're looking for. If multiple queries
   * are given, the first query is the super-query for the bloom
   * filter.
   **/
  HashCode160 queries[0];

} AFS_CS_QUERY;

/**
 * Format of a request.
 **/
typedef struct RequestEntry {

  /**
   * The message that is send to gnunetd.
   **/
  AFS_CS_QUERY * message;
  
  /**
   * Last time the query was send.
   **/
  cron_t lasttime;

  /**
   * Whom to call once we get a reply?
   **/
  Listener receiver;

  /**
   * The node to pass to the receiver method.
   **/
  struct Block * receiverNode;

  /**
   * Opaque data handle to pass to the Listener.
   **/
  void * data;

  /**
   * How long have we been actively trying this one?
   **/
  int tries;

} RequestEntry;

/* ************* context and Block **************** */

/**
 * Progress of the current operation. Used for passing
 * data to callbacks. Some of these make sense only for 
 * downloading.
 **/
typedef struct ProgressStats {

  size_t progress;			/* bytes processed */
  size_t filesize;			/* total file size */
  int requestsSent;
  int requestsPending;
  int currentRetries;
  int totalRetries;
  int currentTTL;
  int duplicationEstimate;

} ProgressStats;


/**
 * Progress model. Called whenever we make progress. Callback methods
 * of this type are used during insertion and download to notify the
 * user interface of the progress we're making. If the model is called
 * with position == total, the download is complete. If the model
 * is called with position == total == 0, then there was a fatal error
 * and the download was aborted.
 * 
 * @param current progress statistics
 * @param data a context passed around for use by the PM
 *        implementation
 **/
typedef void (*ProgressModel)(ProgressStats * stats,
			      void * data);

/**
 * The NodeContext groups the IOC and the progress model
 * into a single struct.
 **/
typedef struct NodeContext {
  
  /**
   * The IO context for IO operations.
   **/
  IOContext ioc;

  /**
   * Priority
   **/
  unsigned int priority;

  /**
   * Index of the file that we are indexing, -1 for insertion
   **/
  unsigned short index;
 
  /**
   * The ProgressModel to communicate status updates.
   **/
  ProgressModel pmodel;

  /**
   * Data argument to the ProgressModel.
   **/
  void * data;

  /**
   * Current progress so far.
   **/
  ProgressStats stats;

} NodeContext;

/**
 * Insert the current block into the network. Implementations
 * are also responsible for updating the corresponding fields
 * of the parent node (of course, except if the parent is
 * NULL in the case of the top-node in the tree).<p>
 *
 * Inner nodes first call the respective inserter methods for
 * their children.<p>
 * 
 * @param this the node that should be inserted or indexed
 * @param ioc the IO context
 * @param sock the socket to use to talk to the core, NULL if
 *        we just do a "fake" insert to compute the tree in memory
 * @return OK on success, SYSERR on error
 **/
typedef int (*Inserter)(void * this,
			NodeContext * nc,
			GNUNET_TCP_SOCKET * sock);

/**
 * Download this node (and the children below). Note that the
 * processing is asynchronous until the pmodel is called with position
 * == total (and thus no more requests are pending) or the request
 * manager is aborted by the user.
 * 
 * @param this the node that should be inserted or indexed
 * @param nc the context
 * @param rm the request manager
 **/
typedef void (*Downloader)(void * this,
			   NodeContext * nc,
			   struct RequestManager * rm);

/**
 * Check if this block is already present, if yes,
 * loads it.
 *
 * @param this the block to check
 * @param nc the context
 * @return YES if the block is present, NO if not
 **/
typedef int (*PresentChecker)(void * this,
			      NodeContext * nc);

/**
 * Print the block summary (for debugging)
 **/
typedef void (*BlockPrinter)(void * this,
			     int ident);

/**
 * Free the associated resources of this Block. DOES ALSO free the
 * memory occupied by the Block struct itself!
 *
 * @param this reference to the Block
 **/
typedef void (*Block_Destructor)(void * this,
				 RequestManager * rm);

/**
 * Shared structure of all blocks (DBlocks and IBlocks).
 **/
typedef struct Block {

  /**
   * The total size of the file.
   **/
  size_t filesize;

  /**
   * Position of the block relative to the beginning of the file.
   **/
  size_t pos;

  /**
   * Hashes of the plaintext block (key) and the encrypted block
   * (query).
   **/
  CHK_Hashes chk;

  /**
   * Free resources of the Block.
   **/
  Block_Destructor done;
  
  /**
   * Insert the block.
   **/
  Inserter insert;

  /**
   * Download the block.
   **/
  Downloader download;

  /**
   * Listener method to receive a reply for the block.
   **/
  Listener receive;

  /**
   * Check if the block is present.
   **/
  PresentChecker isPresent;

  /**
   * Print the node to the logger.
   **/
  BlockPrinter print;

  /**
   * How many bytes in data are actual data (not padding)?
   * Set to 0 to indicate that the download of this block
   * is complete.
   **/
  unsigned int len;

  /**
   * Pointer to the data of this block, NULL if the data is not yet
   * available.
   **/
  void * data;

  /**
   * The parent node in the file-tree, NULL for the node on top of the
   * file-tree.
   **/
  struct IBlock * parent;

  /**
   * CRC of the top-IBlock (not used except for the root of the
   * tree), see downloadutil.c and block.c::childDownloadCompleted.
   **/ 
  int topCrc32;

  /**
   * See BLOCK_XXX constants.
   **/
  short status;

} Block; /* total: 94 bytes, 24 bytes could be saved by using a shared-VTBL 
	    and status-bits to indicate which one is used */

/**
 * Block is freshly created, nothing has been done.
 **/
#define BLOCK_CREATED 0

/**
 * We know the correct block data and is is on the drive
 * (and in memory if data != NULL)
 **/
#define BLOCK_PRESENT 1

/**
 * We do not know the correct data, but we have not done a
 * request yet. It may be that we can construct the data from
 * the children (if they are present).
 **/
#define BLOCK_NOT_PRESENT 2

/**
 * We have a request pending for this block (either with the
 * parent if parent != NULL) or a direct request if 
 * parent == NULL.
 **/
#define BLOCK_PENDING 3

/**
 * This block is present and all children (transitively)
 * are also present.
 **/
#define BLOCK_CHILDREN_PRESENT 4

/**
 * This iblock has a super-query pending.
 **/
#define BLOCK_SUPERQUERY_PENDING 5

/**
 * This block is done (about to be freed).
 **/
#define BLOCK_DONE 6

/**
 * This block shall not be freed, even if all children
 * are dead.
 **/
#define BLOCK_PERSISTENT 7

/* ****************** Leaf struct ************************ */

/**
 * This struct represents a leaf (level-zero node) in the
 * file-tree. 
 **/
typedef struct DBlock {

  /**
   * The shared properties of all types of blocks.
   **/
  Block common;

} DBlock;
		
/**
 * Create a top-DBlock for files <= 1k where there is no parent
 * IBlock. Note that you must set the chk field before calling
 * download.
 *
 * @param filesize the size of the file 
 * @return the DBlock on success, NULL on failure
 **/
Block * createTopDBlock(size_t filesize);
			 
 
/* ******************** Inner nodes ********************* */

/**
 * Number of CHK_Hashes per IBlock. The value must be 25 since
 * 25*40+20+4 is 1024. The other values are 40=sizeof(CHK_Hashes),
 * 20=sizeof(HashCode160) for the super-hash and 4=sizeof(int) for the
 * CRC32.
 **/
#define CHK_PER_INODE 25

/**
 * The plaintext format of an IBlock.
 **/
typedef struct IBlockData {

  /**
   * The super-Hashcode for retrieving all CHK_PER_INODE sub-nodes in one
   * big lookup. This hash is the hash of the concatenation
   * of all encrypted CHK_PER_INODE children of this node.
   **/
  HashCode160 superHash;

  /**
   * The CRC32 checksum of the sub-blocks (crc32N of
   * the concatenation of the individual crc32N's over
   * the plaintext-data (without padding) of each block).
   **/
  int crc32;

  /**
   * The keys and queries for the nodes one level below.
   * This entry must be at the end since it is variable
   * size!
   **/
  CHK_Hashes chks[CHK_PER_INODE];

} IBlockData;

/**
 * This struct represents an IBlock (inner node) in the
 * file-tree. 
 **/
typedef struct IBlock {

  /**
   * The shared properties of all types of blocks.
   **/
  Block common;

  /**
   * The depth of this node in the file tree.
   * At depth 0 we have the leaves, since this is
   * an IBlock, depth is always > 0.
   **/
  unsigned int depth;

  /**
   * Number of children [1-CHK_PER_INODE] of this node.
   **/
  unsigned int childcount;

  /**
   * References to the children (IBlocks or DBlocks, 
   * depending on if depth > 1 or not).
   **/
  Block * children[CHK_PER_INODE];

  /**
   * CRC of each of the children.
   **/
  int crcs[CHK_PER_INODE];

  /**
   * Pointer to the parent IBlock.
   **/
  struct IBlock * parent;

} IBlock;

/**
 * Create an IBlock. Use createTopIBlock for the
 * node on top of the file-tree.
 *
 * @param pos the position of the IBlock in the file
 * @param parent the parent block (may not be NULL)
 * @return the IBlock 
 **/
Block * createIBlock(size_t pos,
		     IBlock * parent);

/**
 * Create a DBlock. Note that this method can NOT be used for files <=
 * 1k since parent may not be NULL (which it would be for the
 * top-block). Use createTopDBlock for files <= 1k.
 *
 * @param pos the offset of this block in the file
 * @param parent the parent block
 * @return the DBlock on success, NULL on failure
 **/
Block * createDBlock(size_t pos,
		     IBlock * parent);
	
/**
 * Create a top-IBlock for the root of the file tree.
 * Note that you must set the chk field before calling
 * download. 
 * @param filesize the size of the file
 **/
Block * createTopIBlock(size_t filesize);

/**
 * A child has been completely downloaded. Perform the
 * appropriate checks in the parent node and free associated
 * resources if possible. Since the only errors are either
 * bugs or hash-crc-collisions (probability 1:2^160), we
 * always die on errors (return values do not work well for
 * async calls anyway).<p>
 * 
 * Note that the leaves update the ProgressModel, so we do
 * not have to worry about that.
 *
 * @param this the IBlock
 * @param child the completed child block
 **/
void childDownloadCompleted(struct IBlock * parent,
			    Block * child,
			    NodeContext * nc,
			    RequestManager * rm);


/* ***************** the root **************************** */

#define MAX_DESC_LEN     256
#define MAX_FILENAME_LEN 128
#define MAX_MIMETYPE_LEN 128
#define ROOT_MINOR_VERSION 0
#define ROOT_MAJOR_VERSION 1

/**
 * A FileIdentifier groups the information
 * required to download (and check) a file.
 **/
typedef struct {
  
  /**
   * Total size of the file in bytes. (network byte order (!)) 
   **/
  size_t file_length;

  /**
   * Top CRC of the tree-encoding. (network byte order (!)) 
   **/
  int crc;

  /**
   * Query and key of the top IBlock.
   **/
  CHK_Hashes chk;
} FileIdentifier;

/**
 * The structure of the root node - contains pertinent information for
 * the file (file length, checksum, hashcode of main indirection node,
 * description length, and description.
 **/
typedef struct {
   
  /**
   * Major format version, in network byte order 
   **/
  unsigned short major_formatVersion;
  
  /**
   * Minor format version, in network byte order 
   **/
  unsigned short minor_formatVersion;
  
  /**
   * Information required for the download.
   **/
  FileIdentifier fileIdentifier;

  /**
   * description of the contents, padded with zeros.
   **/
  char description[MAX_DESC_LEN];

  /**
   * suggested filename, padded with zeros. 
   **/
  char filename[MAX_FILENAME_LEN];
  
  /**
   * mime-type (as claimed by insertion!) 
   **/
  char mimetype[MAX_MIMETYPE_LEN];

} RootNodeHeader;

/**
 * The structure of the root node, including padding to make
 * it to 1k.
 **/
typedef struct {

  /**
   * The real data in the root-node 
   **/
  RootNodeHeader header;

  /**
   * Padding 
   **/
  char padding[CONTENT_SIZE - sizeof(RootNodeHeader)];

} RootNode;


#endif
