/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file transports/tcp.c
 * @author Christian Grothoff
 *
 * Implementation of the TCP transport layer.
 **/

#include "transports/tcp.h"
#include "util/ipcheck.h"
#include "util/semaphore.h"
#include "util/logging.h"
#include "util/ports.h"
#include "util/identity.h"
#include "util/configuration.h"
#include "util/statistics.h"
#include "util/statuscalls.h"
#include "util/tcpio.h"
#include <errno.h>
#include <unistd.h>

/**
 * after how much time of the core not being associated with a tcp
 * connection anymore do we close it? 
 **/
#define TCP_TIMEOUT 30 * cronSECONDS

/**
 * Host-Address in a TCP network.
 **/
typedef struct {
  /**
   * claimed IP of the sender, network byte order 
   **/  
  IPaddr ip;

  /**
   * claimed port of the sender, network byte order 
   **/
  unsigned short port; 

  /**
   * reserved (set to 0 for signature verification) 
   **/
  unsigned short reserved; 

} HostAddress;

/**
 * TCP Message-Packet header. 
 **/
typedef struct {
  /**
   * size of the message, in bytes, including this header; 
   * max 65536-header (network byte order) 
   **/
  unsigned short size;

  /**
   * reserved, must be 0 (network byte order) 
   **/
  unsigned short isEncrypted;

  /**
   * CRC checksum of the packet  (network byte order)
   **/ 
  int checkSum;
  
  /**
   * This struct is followed by MESSAGE_PARTs - until size is reached 
   * There is no "end of message".
   **/
  p2p_HEADER parts[0];
} TCPMessagePack;

/**
 * Initial handshake message. Note that the beginning
 * must match the CS_HEADER since we are using tcpio.
 **/
typedef struct {
  /**
   * size of the handshake message, in nbo, value is 24 
   **/    
  unsigned short size;

  /**
   * "message type", TCP version number, always 0.
   **/
  unsigned short version;

  /**
   * Identity of the node connecting (TCP client) 
   **/
  HostIdentity clientIdentity;
} TCPWelcome;

/**
 * Transport Session handle.
 **/
typedef struct {
  /**
   * the tcp socket 
   **/
  int sock;

  /**
   * number of users of this session 
   **/
  int users;

  /**
   * Last time this connection was used
   **/
  cron_t lastUse;

  /**
   * mutex for synchronized access to 'users' 
   **/
  Mutex lock;

  /**
   * To whom are we talking to (set to our identity
   * if we are still waiting for the welcome message)
   **/
  HostIdentity sender;

  /**
   * Are we still expecting the welcome? (YES/NO)
   **/
  int expectingWelcome;

  /**
   * Current read position in the buffer.
   **/
  int pos;

  /**
   * The read buffer.
   **/
  char buff[MAX_BUFFER_SIZE];

  /**
   * Position in the write buffer
   **/
  int wpos;

  /**
   * The write buffer.
   **/
  char * wbuff;

} TCPSession;

/* *********** globals ************* */

/**
 * apis (our advertised API and the core api ) 
 **/
static CoreAPIForTransport * coreAPI;
static TransportAPI tcpAPI;

/**
 * one thread for listening for new connections,
 * and for reading on all open sockets 
 **/
static pthread_t listenThread;

/**
 * sock is the tcp socket that we listen on for new inbound
 * connections.
 **/
static int tcp_sock;

/**
 * tcp_pipe is used to signal the thread that is
 * blocked in a select call that the set of sockets to listen
 * to has changed.
 **/
static int tcp_pipe[2];

/**
 * Array of currently active TCP sessions. 
 **/
static TSession ** tsessions = NULL;
static int tsessionCount;
static int tsessionArrayLength;

/**
 * handles for statistics 
 **/
static int stat_octets_total_tcp_in;
static int stat_octets_total_tcp_out;

/* configuration */
static CIDRNetwork * filteredNetworks_;

/**
 * Lock for access to mutable state of the module,
 * that is the configuration and the tsessions array.
 * Note that we ONLY need to synchronize access to
 * the tsessions array when adding or removing sessions,
 * since removing is done only by one thread and we just
 * need to avoid another thread adding an element at the
 * same point in time. We do not need to synchronize at
 * every access point since adding new elements does not
 * prevent the select thread from operating and removing
 * is done by the only therad that reads from the array.
 **/
static Mutex tcplock;

/**
 * Semaphore used by the server-thread to signal that
 * the server has been started -- and later again to
 * signal that the server has been stopped.
 **/
static Semaphore * serverSignal = NULL;
static int tcp_shutdown = YES;

/* ******************** helper functions *********************** */

/*
 * Catch sigpipe on send, and do nothing
 */
#ifndef LINUX
static void catcher(int sig) {
  LOG(LOG_INFO,
    "INFO: signal %d caught\n", sig);
  signal(sig, catcher);
}
#endif

/**
 * Check if we are allowed to connect to the given IP.
 **/
static int isBlacklisted(struct in_addr ip) {
  int ret;

  MUTEX_LOCK(&tcplock);
  ret = checkIPListed(filteredNetworks_,
		      ip);
  MUTEX_UNLOCK(&tcplock);
  return ret;
}

/**
 * Write to the pipe to wake up the select thread (the set of
 * files to watch has changed).
 **/
static void signalSelect() {
  char i = 0;
  write(tcp_pipe[1],
	&i, 
	sizeof(char));
}

/**
 * Disconnect from a remote node. May only be called
 * on sessions that were aquired by the caller first.
 * For the core, aquiration means to call associate or
 * connect. The number of disconnects must match the
 * number of calls to connect+associate.
 *
 * @param session the session that is closed
 * @return OK on success, SYSERR if the operation failed
 **/
static int tcpDisconnect(TSession * tsession) {
  if (tsession->internal != NULL) {
    TCPSession * tcpsession = tsession->internal;

    MUTEX_LOCK(&tcpsession->lock);
    tcpsession->users--;
    if (tcpsession->users > 0) {
      MUTEX_UNLOCK(&tcpsession->lock);
      return OK;
    }
    MUTEX_UNLOCK(&tcpsession->lock);
    MUTEX_DESTROY(&tcpsession->lock);
    freeNonNull(tcpsession->wbuff,
		"freeing write buffer");
    xfree(tcpsession,
	  "freeing TCPSession (from tcpConnect)");
    xfree(tsession,
	  "freeing TSession (from tcpConnect)");    
  }
  return OK;
}

/**
 * Remove a session, either the other side closed the connection
 * or we have otherwise reason to believe that it should better
 * be killed. Destroy session closes the session as far as the
 * TCP layer is concerned, but since the core may still have
 * references to it, tcpDisconnect may not instantly free all
 * the associated resources.
 *
 * @param i index to the session handle
 **/
static void destroySession(int i) {  
  TCPSession * tcpSession;

  tcpSession = tsessions[i]->internal;
  close(tcpSession->sock);
  tcpSession->sock = -1;
  tcpDisconnect(tsessions[i]);
  MUTEX_LOCK(&tcplock);
  tsessions[i] = tsessions[--tsessionCount];
  tsessions[tsessionCount] = NULL;
  MUTEX_UNLOCK(&tcplock);
}

/**
 * Get the GNUnet UDP port from the configuration,
 * or from /etc/services if it is not specified in 
 * the config file.
 **/
static unsigned short getGNUnetTCPPort() {
  struct servent * pse;	/* pointer to service information entry	*/
  unsigned short port;

  port = (unsigned short) getConfigurationInt("TCP",
					      "PORT");
  if (port == 0) { /* try lookup in services */
    if ((pse = getservbyname("gnunet", "tcp"))) 
      port = htons(pse->s_port);      
  }
  return port;
}

/**
 * A (core) Session is to be associated with a transport session. The
 * transport service may want to know in order to call back on the
 * core if the connection is being closed. Associate can also be
 * called to test if it would be possible to associate the session
 * later, in this case the argument session is NULL. This can be used
 * to test if the connection must be closed by the core or if the core
 * can assume that it is going to be self-managed (if associate
 * returns OK and session was NULL, the transport layer is responsible
 * for eventually freeing resources associated with the tesession). If
 * session is not NULL, the core takes responsbility for eventually
 * calling disconnect.
 * 
 * @param tsession the session handle passed along
 *   from the call to receive that was made by the transport
 *   layer
 * @return OK if the session could be associated,
 *         SYSERR if not.
 **/
static int tcpAssociate(TSession * tsession) {
  TCPSession * tcpSession;

  if (tsession == NULL) {
    LOG(LOG_FAILURE,
	"FAILURE: assertFailed: tcpAssociate called with tsession NULL\n");
    return SYSERR;
  }
  tcpSession = (TCPSession*) tsession->internal;
  MUTEX_LOCK(&tcpSession->lock);
  tcpSession->users++;
  MUTEX_UNLOCK(&tcpSession->lock);
  return OK;
}

/**
 * The socket of session i has data waiting, process!
 **/
static int readAndProcess(int i) {
  TSession * tsession;
  TCPSession * tcpSession;
  int len;
  TCPMessagePack * pack;
  MessagePack * mp;

  tsession = tsessions[i];
  if (SYSERR == tcpAssociate(tsession))
    return SYSERR;
  tcpSession = tsession->internal;
  len = read(tcpSession->sock,
	     &tcpSession->buff[tcpSession->pos],
	     MAX_BUFFER_SIZE - tcpSession->pos);
  cronTime(&tcpSession->lastUse);
  if (len == 0) {
    tcpDisconnect(tsession);
#if DEBUG_TCP
    LOG(LOG_DEBUG,
	"DEBUG: READ on socket %d returned 0 bytes, closing connection\n",
	tcpSession->sock);
#endif
    return SYSERR; /* other side closed connection */
  }
  if (len < 0) {
    if ( (errno == EINTR) ||
	 (errno == EAGAIN) ) { 
#if DEBUG_TCP
      LOG(LOG_DEBUG,
	  "DEBUG: READ on socket %d returned %s, closing connection\n",
	  tcpSession->sock,
	  strerror(errno));
#endif
      tcpDisconnect(tsession);
      return OK;    
    }
    LOG(LOG_INFO,
	"INFO: read failed on peer tcp connection (%d), closing (%s).\n",
	len,
	strerror(errno));
    tcpDisconnect(tsession);
    return SYSERR;
  }
  incrementBytesReceived(len);
  statChange(stat_octets_total_tcp_in,
	     len);
  tcpSession->pos += len;
  len = ntohs(((TCPMessagePack*)&tcpSession->buff[0])->size);
#if DEBUG_TCP
  LOG(LOG_DEBUG,
      "Read %d bytes on socket %d, expecting %d for full message\n",
      tcpSession->pos,
      tcpSession->sock, 
      len);
#endif
  if ( (tcpSession->pos < 2) ||
       (tcpSession->pos < len) ) {
    tcpDisconnect(tsession);
    return OK;
  }
 
  /* complete message received, let's check what it is */
  if (YES == tcpSession->expectingWelcome) {
    TCPWelcome * welcome;
#if DEBUG_TCP
    HexName hex;
#endif
    
    welcome = (TCPWelcome*) &tcpSession->buff[0];
    if ( (ntohs(welcome->version) != 0) ||
	 (ntohs(welcome->size) != sizeof(TCPWelcome)) ) {
      LOG(LOG_WARNING,
	  "WARNING: expected welcome on tcp connection, got garbage (%d, %d). Closing.\n",
	  ntohs(welcome->version),
	  ntohs(welcome->size));
      tcpDisconnect(tsession);
      return SYSERR;
    }
    tcpSession->expectingWelcome = NO;
    memcpy(&tcpSession->sender,
	   &welcome->clientIdentity,
	   sizeof(HostIdentity));     
#if DEBUG_TCP
    IFLOG(LOG_DEBUG,
	  hash2hex(&tcpSession->sender.hashPubKey,
		   &hex));
    LOG(LOG_DEBUG,
	"DEBUG: tcp welcome message from %s received\n",
	&hex);
#endif
    memcpy(&tcpSession->buff[0],
	   &tcpSession->buff[sizeof(TCPWelcome)],
	   tcpSession->pos - sizeof(TCPWelcome));
    tcpSession->pos -= sizeof(TCPWelcome); 
    len = ntohs(((TCPMessagePack*)&tcpSession->buff[0])->size);
  } 
  if ( (tcpSession->pos < 2) ||
       (tcpSession->pos < len) ) {
    tcpDisconnect(tsession);
    return OK;
  }
     
  pack = (TCPMessagePack*)&tcpSession->buff[0];
  /* send msg to core! */
  if (len <= sizeof(TCPMessagePack)) {
    LOG(LOG_WARNING,
	"WARNING: received malformed message from tcp-peer connection. Closing.\n");
    tcpDisconnect(tsession);
    return SYSERR;
  }
  mp      = xmalloc(sizeof(MessagePack), 
		    "tcp allocating message pack");
  mp->msg = xmalloc(len, 
		    "tcp allocating buffer for message");
  memcpy(mp->msg,
	 &pack->parts[0],
	 len - sizeof(TCPMessagePack));
  memcpy(&mp->sender,
	 &tcpSession->sender,
	 sizeof(HostIdentity));
  mp->crc         = ntohl(pack->checkSum);
  mp->isEncrypted = ntohs(pack->isEncrypted);
  mp->size        = len - sizeof(TCPMessagePack);
  mp->tsession    = tsession;
#if DEBUG_TCP
  LOG(LOG_DEBUG,
      "DEBUG: tcp transport received %d bytes, forwarding to core\n",
      mp->size);
#endif
  coreAPI->receive(mp);

  if (tcpSession->pos < len) { 
    LOG(LOG_FAILURE,
	"FAILURE: assert failed, pos (%d) < len (%d)\n",
	tcpSession->pos, len);
    tcpDisconnect(tsession);
    return SYSERR;
  }
  /* finally, shrink buffer adequately */
  memcpy(&tcpSession->buff[0],
	 &tcpSession->buff[len],
	 tcpSession->pos - len);
  tcpSession->pos -= len;	   
  
  tcpDisconnect(tsession);
  return OK;
}

/**
 * Add a new session to the array watched by the select thread.
 * Grows the array if needed. 
 **/
static void addTSession(TSession * tsession) {
  MUTEX_LOCK(&tcplock);
  if (tsessionCount == tsessionArrayLength) {
    TSession ** tmp;
    tmp = xmalloc(tsessionArrayLength*2*sizeof(TSession*),
		  "growing tsession array");
    memcpy(tmp, tsessions, sizeof(TSession*) * tsessionArrayLength);
    tsessionArrayLength *= 2;    
    xfree(tsessions, 
	  "freeing old tsession array");
    tsessions = tmp;
  }
  tsessions[tsessionCount++] = tsession;
  MUTEX_UNLOCK(&tcplock);
}

/**
 * Create a new session for an inbound connection on the given
 * socket. Adds the session to the array of sessions watched
 * by the select thread.
 **/
static void createNewSession(int sock) {
  TSession * tsession;
  TCPSession * tcpSession;

  tcpSession = xmalloc(sizeof(TCPSession), 
		       "tcpsession in handleConnection");
  tcpSession->pos = 0;
  tcpSession->wpos = 0;
  tcpSession->wbuff = NULL;
  tcpSession->sock = sock;
  /* fill in placeholder identity to mark that we 
     are waiting for the welcome message */
  memcpy(&tcpSession->sender,
	 coreAPI->myIdentity,
	 sizeof(HostIdentity));
  tcpSession->expectingWelcome = YES;
  MUTEX_CREATE(&tcpSession->lock);
  tcpSession->users = 1; /* us only, core has not seen this tsession! */
  cronTime(&tcpSession->lastUse);
  tsession = xmalloc(sizeof(TSession),
		     "tsession in handleConnection");
  tsession->ttype = TCP_PROTOCOL_NUMBER;
  tsession->internal = tcpSession;
  addTSession(tsession);
}					 

/**
 * Main method for the thread listening on the tcp socket and all tcp
 * connections. Whenever a message is received, it is forwarded to the
 * core. This thread waits for activity on any of the TCP connections
 * and processes deferred (async) writes and buffers reads until an
 * entire message has been received.
 **/
static void tcpListenMain() {
  struct sockaddr_in clientAddr;
  fd_set readSet;
  fd_set errorSet;
  fd_set writeSet;
  int lenOfIncomingAddr;
  int i;
  int max;
  int ret;

  if (tcp_sock != -1)
    listen(tcp_sock, 5);
  semaphore_up(serverSignal); /* we are there! */
  while (tcp_shutdown == NO) {
    FD_ZERO(&readSet);
    FD_ZERO(&errorSet);
    FD_ZERO(&writeSet);
    if (tcp_sock != -1) 
      FD_SET(tcp_sock, &readSet);
    FD_SET(tcp_pipe[0], &readSet);
    max = tcp_pipe[0];
    if (tcp_sock > tcp_pipe[0])
      max = tcp_sock;
    for (i=0;i<tsessionCount;i++) {
      TCPSession * tcpSession = tsessions[i]->internal;
      int sock = tcpSession->sock;
      FD_SET(sock, &readSet);
      FD_SET(sock, &errorSet);
      if (tcpSession->wpos > 0)
	FD_SET(sock, &writeSet); /* do we have a pending write request? */
      if (sock > max)
	max = sock;
    }    
    ret = select(max+1, &readSet, &writeSet, &errorSet, NULL);    
    if ( (ret == -1) &&
	 ( (errno == EAGAIN) || (errno == EINTR) ) ) 
      continue;    
    if (ret == -1) {
      if (errno == EBADF) {
	LOG(LOG_WARNING,
	    "WARNING: %s in select, trying to recover (bug!)\n",
	    strerror(errno));
	sleep(1); /* sleep for 1s to avoid looping too quick
		     if the error should be persistent! */
	continue;
      }
      errexit("FAILURE in select: %s\n",
	      strerror(errno));
      break; /* DIE */
    }
    if (tcp_sock != -1) {
      if (FD_ISSET(tcp_sock, &readSet)) {
	int sock;
	
	lenOfIncomingAddr = sizeof(clientAddr);               
	sock = accept(tcp_sock, 
		      (struct sockaddr *)&clientAddr, 
		      &lenOfIncomingAddr);
	/* verify clientAddr for eligibility here (ipcheck-style,
	   user should be able to specify who is allowed to connect,
	   otherwise we just close and reject the communication! */  
	if (YES == isBlacklisted(clientAddr.sin_addr)) {
	  LOG(LOG_INFO,
	      "INFO: Rejected blacklisted connection from %d.%d.%d.%d.\n",
	      PRIP(ntohl(*(int*)&clientAddr.sin_addr)));
	  close(sock);
	} else 
	  createNewSession(sock);      
      }
    }
    if (FD_ISSET(tcp_pipe[0], &readSet)) 
      read(tcp_pipe[0], &i, 1); /* just a signal to refresh sets, eat and continue */
    for (i=0;i<tsessionCount;i++) {
      TCPSession * tcpSession = tsessions[i]->internal;
      int sock = tcpSession->sock;
      if (FD_ISSET(sock, &readSet)) {
	if (SYSERR == readAndProcess(i)) {
	  destroySession(i);
	  i--;
	  continue;
	}
      }
      if (FD_ISSET(sock, &writeSet)) {
	int ret;
#ifndef LINUX
  if ( SIG_ERR == signal(SIGPIPE, SIG_IGN)) {
    if (SIG_ERR == signal(SIGPIPE, catcher))
      LOG(LOG_WARNING,
        "WARNING: could not install handler for SIGPIPE!\n"\
        "Attempting to continue anyway.");
  }
	ret = send(sock,
		   tcpSession->wbuff,
		   tcpSession->wpos,
		   MSG_DONTWAIT);
#else
	ret = send(sock,
		   tcpSession->wbuff,
		   tcpSession->wpos,
		   MSG_DONTWAIT|MSG_NOSIGNAL);
#endif
	if (ret == -1) {
	  destroySession(i);
	  i--;
	  continue;
	}
	if (ret == tcpSession->wpos) {
	  xfree(tcpSession->wbuff,
		"freeing write buffer");
	  tcpSession->wbuff = NULL;
	  tcpSession->wpos = 0;
	} else {
	  memcpy(tcpSession->wbuff,
		 &tcpSession->wbuff[ret],
		 tcpSession->wpos - ret);
	  tcpSession->wpos -= ret;
	}
      }
      if (FD_ISSET(sock, &errorSet)) {
	destroySession(i);
	i--;
	continue;
      }
      if ( ( tcpSession->users == 1) &&
	   (cronTime(NULL) > tcpSession->lastUse + TCP_TIMEOUT) ) {
	destroySession(i);
	i--;
	continue;
      }
    }
  }
  /* shutdown... */
  if (tcp_sock != -1) {
    close(tcp_sock);
    tcp_sock = -1;
  }
  /* close all sessions */
  while (tsessionCount > 0) 
    destroySession(0);    
  semaphore_up(serverSignal); /* we are there! */
} /* end of tcp listen main */

/**
 * Send a message (already encapsulated if needed) via the
 * tcp socket (or enqueue if sending now would block).
 *
 * @param tcpSession the session to use for sending
 * @param mp the message to send
 * @param ssize the size of the message
 * @return OK if message send or queued, SYSERR if queue is full and
 * message was dropped.
 **/
static int tcpDirectSend(TCPSession * tcpSession,
			 void * mp,
			 int ssize) {
  int ok;
  int ret;

  if (tcpSession->sock == -1) {
    LOG(LOG_INFO,
	"INFO: tcpDirectSend called, but socket is closed\n");
    return SYSERR;
  }
  ok = SYSERR;
  if (tcpSession->wpos > 0) 
    ret = 0;
  else {
#ifndef LINUX
    if ( SIG_ERR == signal(SIGPIPE, SIG_IGN)) {
      if (SIG_ERR == signal(SIGPIPE, catcher))
        LOG(LOG_WARNING,
          "WARNING: could not install handler for SIGPIPE!\n"\
          "Attempting to continue anyway.");
    }
    ret = send(tcpSession->sock,
	       mp,
	       ssize,
	       MSG_DONTWAIT);
#else
    ret = send(tcpSession->sock,
	       mp,
	       ssize,
	       MSG_DONTWAIT|MSG_NOSIGNAL);
#endif
  }
  if (ret == -1) {
    if ( (errno == EAGAIN) || (errno == EWOULDBLOCK))
      ret = 0;
    else {
      LOG(LOG_INFO,
	      "INFO: write to tcp peer failed (%s)\n",
	    strerror(errno));
      return SYSERR;
    }
  }
  if ( (ret>=0) && (ret <= ssize) ) { /* some bytes send or blocked */
    if (ret < ssize) {
      if (tcpSession->wbuff == NULL) 
	tcpSession->wbuff = xmalloc(tcpAPI.mtu + sizeof(TCPMessagePack),
				    "allocating write buffer");
      if (ssize - ret > tcpAPI.mtu + sizeof(TCPMessagePack) - tcpSession->wpos)
	ok = SYSERR; /* buffer full, drop */
      else {
	memcpy(&tcpSession->wbuff[tcpSession->wpos],
	       mp,
	       ssize - ret);
	tcpSession->wpos += ssize - ret;
	if (tcpSession->wpos > ssize - ret)
	  signalSelect(); /* select set changed! */
	ok = OK; /* all buffered */
      }      
    } else 
      ok = OK; /* all written */
  } else {
    LOG(LOG_WARNING, 
	"WARNING: send failed (%s) - %d %d\n",
	strerror(errno),
	errno,
	ret);
    ok = SYSERR; /* write failed for real */
  }
  cronTime(&tcpSession->lastUse);
  incrementBytesSent(ssize);
  statChange(stat_octets_total_tcp_out,
	     ssize);
  return ok;
}

/**
 * Verify that a HELO-Message is correct (a node
 * is reachable at that address). Since the reply
 * will be asynchronous, a method must be called on
 * success. 
 * @param helo the HELO message to verify
 *        (the signature/crc have been verified before)
 * @param callback the method to call on success
 * @param xarg an opaque argument to pass to the callback
 *        method (typically == helo, but not guaranteed)
 **/
static int verifyHelo(HELO_Message * helo) {
  HostAddress * haddr;

  haddr = (HostAddress*) &helo->senderAddress[0];
  if ( (ntohs(helo->senderAddressSize) != sizeof(HostAddress)) ||
       (ntohs(helo->header.size) != HELO_Message_size(helo)) ||
       (ntohs(helo->header.requestType) != p2p_PROTO_HELO) ||
       (ntohs(helo->protocol) != TCP_PROTOCOL_NUMBER) ||
       (YES == isBlacklisted(haddr->ip.addr)) )
    return SYSERR; /* obviously invalid */
  else
    return OK;
}

/**
 * Create a HELO-Message for the current node. The HELO is
 * created without signature and without a timestamp. The
 * GNUnet core will sign the message and add an expiration time. 
 *
 * @param identity the HostIdentity of the current node
 * @param address the address of the current node
 * @param helo address where to store the pointer to the HELO
 *        message
 * @return OK on success, SYSERR on error
 **/
static int createHELO(HELO_Message ** helo) {
  HELO_Message * msg;
  HostAddress * haddr;
  unsigned short port;

  port = getGNUnetTCPPort();
  if (0 == port)
    return SYSERR; /* TCP transport is configured SEND-only! */
  msg = (HELO_Message *) xmalloc(sizeof(HELO_Message) + sizeof(HostAddress),
				 "creating HELO (tcp)");
  haddr = (HostAddress*) &msg->senderAddress[0];

  if (SYSERR == getPublicIPAddress(&haddr->ip)) {
    xfree(msg,
	  "helo creation failed, freeing allocated space");
    LOG(LOG_WARNING,
	"WARNING: Could not determine my public IP address.\n");
    return SYSERR;
  }
  haddr->port = htons(port); 
  haddr->reserved = htons(0);
  msg->senderAddressSize = htons(sizeof(HostAddress));
  msg->protocol = htons(TCP_PROTOCOL_NUMBER);
  msg->MTU = htonl(tcpAPI.mtu);
  *helo = msg;
  return OK;
}

/**
 * Establish a connection to a remote node.
 * @param helo the HELO-Message for the target node
 * @param session the session handle that is set
 * @return OK on success, SYSERR if the operation failed
 **/
static int tcpConnect(HELO_Message * helo,
		      TSession ** tsessionPtr) {
  int i;
  HostAddress * haddr;
  TCPWelcome welcome;
  int sock;
  TSession * tsession;
  TCPSession * tcpSession;
  struct sockaddr_in soaddr;

  haddr = (HostAddress*) &helo->senderAddress[0];
  LOG(LOG_DEBUG,
      "DEBUG: creating TCP connection to %d.%d.%d.%d:%d\n",
      PRIP(ntohl(*(int*)&haddr->ip.addr)), 
      ntohs(haddr->port));
  sock = socket(PF_INET, SOCK_STREAM, 6);/* 6: TCP */
  if (sock == -1) {
    LOG(LOG_FAILURE,
	"FAILURE: Can not create socket (%s).\n",
	strerror(errno));
    return SYSERR;
  }
  if (0 != fcntl(sock, F_SETFL, O_NONBLOCK)) {
    close(sock);
    LOG(LOG_FAILURE,
	"FAILURE: could not put tcp socket into non-blocking mode (%s)\n",
	strerror(errno));
    return SYSERR;
  }
  soaddr.sin_family = AF_INET;
  soaddr.sin_addr = haddr->ip.addr;
  soaddr.sin_port = haddr->port;
  i = connect(sock, 
	      (struct sockaddr*)&soaddr,
	      sizeof(soaddr));
  if ( (i < 0) &&
       (errno != EINPROGRESS) ) {
    LOG(LOG_ERROR,
	"ERROR: Can not connect to %d.%d.%d.%d:%d (%s)\n",
	PRIP(ntohl(*(int*)&haddr->ip)),
	ntohs(haddr->port),
	strerror(errno));
    close(sock);
    return SYSERR;
  }  
  tcpSession = xmalloc(sizeof(TCPSession),
		       "tcp connect");
  tcpSession->sock = sock;
  tcpSession->wpos = 0;
  tcpSession->wbuff = NULL;
  tsession = xmalloc(sizeof(TSession),
		     "tcp connect");   
  tsession->internal = tcpSession;
  tsession->ttype = tcpAPI.protocolNumber;
  MUTEX_CREATE(&tcpSession->lock);
  tcpSession->users = 2; /* caller + us */
  tcpSession->pos = 0;
  cronTime(&tcpSession->lastUse);
  memcpy(&tcpSession->sender,
	 &helo->senderIdentity,
	 sizeof(HostIdentity));
  tcpSession->expectingWelcome = NO;
  addTSession(tsession);
  signalSelect();

  /* send our node identity to the other side to fully establish the
     connection! */
  welcome.size = htons(sizeof(TCPWelcome));
  welcome.version = htons(0);
  memcpy(&welcome.clientIdentity,
	 coreAPI->myIdentity,
	 sizeof(HostIdentity));
  tcpDirectSend(tcpSession,
		&welcome,
		sizeof(TCPWelcome));
  *tsessionPtr = tsession;
  xfree(helo, "freeing helo, connect successful");
  return OK;
}

/**
 * Send a message to the specified remote node.
 * @param session the HELO_Message identifying the remote node
 * @param msg the message
 * @param size the size of the message
 * @return SYSERR on error, OK on success
 **/
static int tcpSend(TSession * tsession,
		   const void * msg,
		   const unsigned int size,
		   int isEncrypted,
		   const int crc) {
  TCPMessagePack * mp;
  int ok;
  int ssize;

  if (size > tcpAPI.mtu) {
    LOG(LOG_FAILURE,
	"FAILURE: message larger than allowed by tcp transport (%d > %d)\n",
	size, 
	tcpAPI.mtu);
    return SYSERR;
  }
  if (((TCPSession*)tsession->internal)->sock == -1)
    return SYSERR; /* other side closed connection */
  mp = xmalloc(sizeof(TCPMessagePack) + size,
	       "allocating space for outbound message");
  mp->checkSum = htonl(crc);
  mp->isEncrypted = htons(isEncrypted);
  memcpy(&mp->parts[0],
	 msg,
	 size);
  ssize = size + sizeof(TCPMessagePack);
  mp->size = htons(ssize);
  
  ok = tcpDirectSend(tsession->internal,
		     mp,
		     ssize);
  xfree(mp,
	"tcp message is out");
  return ok;
}

/**
 * Start the server process to receive inbound traffic.
 * @return OK on success, SYSERR if the operation failed
 **/
static int startTransportServer(void) {
  struct sockaddr_in serverAddr;
  const int on = 1;
  unsigned short port;
  
  if (serverSignal != NULL) {
    LOG(LOG_FAILURE,
	"FAILURE: can not start TCP server, already running!?\n");
    return SYSERR;
  }
  serverSignal = new_semaphore(0);
  tcp_shutdown = NO;
    
  if (0 != pipe(tcp_pipe)) {
    LOG(LOG_ERROR,
	"ERROR: could not create pipe (%s)\n",
	strerror(errno));
    return SYSERR;
  }
  tsessionCount = 0;
  tsessionArrayLength = getConfigurationInt("TCP",
					    "MAXTCPCONNECTS");
  tsessions = xmalloc(sizeof(TSession*) * tsessionArrayLength,
		     "allocating initial tsession array");
  
  port = getGNUnetTCPPort();
  if (port != 0) { /* if port == 0, this is a read-only
		      business! */
    tcp_sock = socket(PF_INET, SOCK_STREAM, 0);
    if (tcp_sock < 0) 
      errexit("ERROR opening tcp socket (%s).\n",
	      strerror(errno));  
    if ( setsockopt(tcp_sock,
		    SOL_SOCKET, 
		    SO_REUSEADDR, 
		    &on, sizeof(on)) < 0 ) 
      errexit("ERROR: setsockopt for tcp socket failed (%s)\n",
	      strerror(errno));  
    memset((char *) &serverAddr, 
	   0,
	   sizeof(serverAddr));
    serverAddr.sin_family      = AF_INET;
    serverAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    serverAddr.sin_port        = htons(getGNUnetTCPPort());
    LOG(LOG_INFO,
	"INFO: starting tcp peer server on port %d\n",
	ntohs(serverAddr.sin_port));
    if (bind(tcp_sock, 
	     (struct sockaddr *) &serverAddr,
	     sizeof(serverAddr)) < 0) {
      LOG(LOG_ERROR, 
	  "ERROR (%s) binding the TCP listener to port %d. "\
	  "No transport service started.\n",
	  strerror(errno),
	  getGNUnetTCPPort());
      close(tcp_sock);
      return SYSERR;
    }
  } else
    tcp_sock = -1;
  PTHREAD_CREATE(&listenThread, 
		 (void * (*)(void *)) &tcpListenMain,
		 NULL,
		 2048);
  semaphore_down(serverSignal); /* wait for server to be up */
  pthread_detach(listenThread);
  addCronJob((CronJob)&signalSelect,
	     5 * cronSECONDS,
	     5 * cronSECONDS,
	     NULL); /* abort select call every 5 seconds to allow
		       timing out connections even if there is
		       no traffic at all */
  return OK;
}

/**
 * Shutdown the server process (stop receiving inbound
 * traffic). Maybe restarted later!
 **/
static int stopTransportServer() {
  tcp_shutdown = YES;  
  signalSelect();
  semaphore_down(serverSignal);
  free_semaphore(serverSignal);
  serverSignal = NULL; 
  delCronJob((CronJob)&signalSelect,
	     5 * cronSECONDS,
	     NULL); 
  close(tcp_pipe[1]);
  close(tcp_pipe[0]);
  if (tcp_sock != -1) 
    close(tcp_sock);
  xfree(tsessions, 
	"freeing tsession array");
  tsessions = NULL;
  tsessionArrayLength = 0;
  return OK;
}

/**
 * Reload the configuration. Should never fail (keep old
 * configuration on error, syslog errors!)
 **/
static void reloadConfiguration(void) {
  char * ch;

  MUTEX_LOCK(&tcplock);
  freeNonNull(filteredNetworks_, 
	      "old blacklist configuration");
  ch = getConfigurationString("TCP",
			      "BLACKLIST");
  if (ch == NULL)
    filteredNetworks_ = parseRoutes("");
  else {
    filteredNetworks_ = parseRoutes(ch);
    xfree(ch, 
	  "network filter computed (reloadConfiguration_tcp)");
  }
  MUTEX_UNLOCK(&tcplock);
}

/**
 * Convert TCP address to a string.
 **/
static char * addressToString(HELO_Message * helo) {
  char * ret;
  HostAddress * haddr;
  
  haddr = (HostAddress*) &helo->senderAddress[0];  
  ret = xmalloc(4*4+6+6,
		"buffer for helo-toString");
  sprintf(ret,
	  "%d.%d.%d.%d:%d (TCP)",
	  PRIP(ntohl(*(int*)&haddr->ip.addr)), 
	  ntohs(haddr->port));
  return ret;
}

 
/* ******************** public API ******************** */
 
/**
 * The exported method. Makes the core api available
 * via a global and returns the udp transport API.
 **/ 
TransportAPI * inittransport_tcp(CoreAPIForTransport * core) {
  int mtu;

  MUTEX_CREATE(&tcplock);
  reloadConfiguration();
  coreAPI = core;
  stat_octets_total_tcp_in 
    = statHandle("# bytes received via tcp");
  stat_octets_total_tcp_out 
    = statHandle("# bytes sent via tcp");
  mtu = getConfigurationInt("TCP",
			    "MTU");
  if (mtu == 0)
    mtu = 1460;
  if (mtu < 1200)
    LOG(LOG_ERROR,
	"ERROR: MTU for TCP is probably to low (fragmentation not implemented!)\n");
 
  tcpAPI.protocolNumber       = TCP_PROTOCOL_NUMBER;
  tcpAPI.mtu                  = mtu - sizeof(TCPMessagePack);
  tcpAPI.cost                 = 20000; /* about equal to udp */
  tcpAPI.verifyHelo           = &verifyHelo;
  tcpAPI.createHELO           = &createHELO;
  tcpAPI.connect              = &tcpConnect;
  tcpAPI.associate            = &tcpAssociate;
  tcpAPI.send                 = &tcpSend;
  tcpAPI.disconnect           = &tcpDisconnect;
  tcpAPI.startTransportServer = &startTransportServer;
  tcpAPI.stopTransportServer  = &stopTransportServer;
  tcpAPI.reloadConfiguration  = &reloadConfiguration;
  tcpAPI.addressToString      = &addressToString;

  return &tcpAPI;
}

void donetransport_tcp() {
  freeNonNull(filteredNetworks_,
	      "tcp blacklist configuration");
  MUTEX_DESTROY(&tcplock);
}

/* end of tcp.c */
